/* 
 * Copyright © 2012 Intel Corporation
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef TEXT_CLIENT_PROTOCOL_H
#define TEXT_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct text_model;
struct text_model_factory;

extern const struct wl_interface text_model_interface;
extern const struct wl_interface text_model_factory_interface;

#ifndef TEXT_MODEL_CONTENT_HINT_ENUM
#define TEXT_MODEL_CONTENT_HINT_ENUM
/**
 * text_model_content_hint - content hint
 * @TEXT_MODEL_CONTENT_HINT_NONE: no special behaviour
 * @TEXT_MODEL_CONTENT_HINT_DEFAULT: auto completion, correction and
 *	capitalization
 * @TEXT_MODEL_CONTENT_HINT_PASSWORD: hidden and sensitive text
 * @TEXT_MODEL_CONTENT_HINT_AUTO_COMPLETION: suggest word completions
 * @TEXT_MODEL_CONTENT_HINT_AUTO_CORRECTION: suggest word corrections
 * @TEXT_MODEL_CONTENT_HINT_AUTO_CAPITALIZATION: switch to uppercase
 *	letters at the start of a sentence
 * @TEXT_MODEL_CONTENT_HINT_LOWERCASE: prefer lowercase letters
 * @TEXT_MODEL_CONTENT_HINT_UPPERCASE: prefer uppercase letters
 * @TEXT_MODEL_CONTENT_HINT_TITLECASE: prefer casing for titles and
 *	headings (can be language dependend)
 * @TEXT_MODEL_CONTENT_HINT_HIDDEN_TEXT: characters should be hidden
 * @TEXT_MODEL_CONTENT_HINT_SENSITIVE_DATA: typed text should not be
 *	stored
 * @TEXT_MODEL_CONTENT_HINT_LATIN: just latin characters should be
 *	entered
 * @TEXT_MODEL_CONTENT_HINT_MULTILINE: the text input is multiline
 *
 * Content hint is a bitmask to allow to modify the behavior of the text
 * input
 */
enum text_model_content_hint {
	TEXT_MODEL_CONTENT_HINT_NONE = 0x0,
	TEXT_MODEL_CONTENT_HINT_DEFAULT = 0x7,
	TEXT_MODEL_CONTENT_HINT_PASSWORD = 0xc,
	TEXT_MODEL_CONTENT_HINT_AUTO_COMPLETION = 0x1,
	TEXT_MODEL_CONTENT_HINT_AUTO_CORRECTION = 0x2,
	TEXT_MODEL_CONTENT_HINT_AUTO_CAPITALIZATION = 0x4,
	TEXT_MODEL_CONTENT_HINT_LOWERCASE = 0x8,
	TEXT_MODEL_CONTENT_HINT_UPPERCASE = 0x10,
	TEXT_MODEL_CONTENT_HINT_TITLECASE = 0x20,
	TEXT_MODEL_CONTENT_HINT_HIDDEN_TEXT = 0x40,
	TEXT_MODEL_CONTENT_HINT_SENSITIVE_DATA = 0x80,
	TEXT_MODEL_CONTENT_HINT_LATIN = 0x100,
	TEXT_MODEL_CONTENT_HINT_MULTILINE = 0x200,
};
#endif /* TEXT_MODEL_CONTENT_HINT_ENUM */

#ifndef TEXT_MODEL_CONTENT_PURPOSE_ENUM
#define TEXT_MODEL_CONTENT_PURPOSE_ENUM
/**
 * text_model_content_purpose - content purpose
 * @TEXT_MODEL_CONTENT_PURPOSE_NORMAL: default input, allowing all
 *	characters
 * @TEXT_MODEL_CONTENT_PURPOSE_ALPHA: allow only alphabetic characters
 * @TEXT_MODEL_CONTENT_PURPOSE_DIGITS: allow only digits
 * @TEXT_MODEL_CONTENT_PURPOSE_NUMBER: input a number (including decimal
 *	separator and sign)
 * @TEXT_MODEL_CONTENT_PURPOSE_PHONE: input a phone number
 * @TEXT_MODEL_CONTENT_PURPOSE_URL: input an URL
 * @TEXT_MODEL_CONTENT_PURPOSE_EMAIL: input an email address
 * @TEXT_MODEL_CONTENT_PURPOSE_NAME: input a name of a person
 * @TEXT_MODEL_CONTENT_PURPOSE_PASSWORD: input a password (combine with
 *	password or sensitive_data hint)
 * @TEXT_MODEL_CONTENT_PURPOSE_DATE: input a date
 * @TEXT_MODEL_CONTENT_PURPOSE_TIME: input a time
 * @TEXT_MODEL_CONTENT_PURPOSE_DATETIME: input a date and time
 * @TEXT_MODEL_CONTENT_PURPOSE_TERMINAL: input for a terminal
 *
 * The content purpose allows to specify the primary purpose of a text
 * input.
 *
 * This allows an input method to show special purpose input panels with
 * extra characters or to disallow some characters.
 */
enum text_model_content_purpose {
	TEXT_MODEL_CONTENT_PURPOSE_NORMAL = 0,
	TEXT_MODEL_CONTENT_PURPOSE_ALPHA = 1,
	TEXT_MODEL_CONTENT_PURPOSE_DIGITS = 2,
	TEXT_MODEL_CONTENT_PURPOSE_NUMBER = 3,
	TEXT_MODEL_CONTENT_PURPOSE_PHONE = 4,
	TEXT_MODEL_CONTENT_PURPOSE_URL = 5,
	TEXT_MODEL_CONTENT_PURPOSE_EMAIL = 6,
	TEXT_MODEL_CONTENT_PURPOSE_NAME = 7,
	TEXT_MODEL_CONTENT_PURPOSE_PASSWORD = 8,
	TEXT_MODEL_CONTENT_PURPOSE_DATE = 9,
	TEXT_MODEL_CONTENT_PURPOSE_TIME = 10,
	TEXT_MODEL_CONTENT_PURPOSE_DATETIME = 11,
	TEXT_MODEL_CONTENT_PURPOSE_TERMINAL = 12,
};
#endif /* TEXT_MODEL_CONTENT_PURPOSE_ENUM */

#ifndef TEXT_MODEL_PREEDIT_STYLE_ENUM
#define TEXT_MODEL_PREEDIT_STYLE_ENUM
enum text_model_preedit_style {
	TEXT_MODEL_PREEDIT_STYLE_DEFAULT = 1,
	TEXT_MODEL_PREEDIT_STYLE_ACTIVE = 2,
	TEXT_MODEL_PREEDIT_STYLE_INACTIVE = 3,
	TEXT_MODEL_PREEDIT_STYLE_HIGHLIGHT = 4,
	TEXT_MODEL_PREEDIT_STYLE_UNDERLINE = 5,
	TEXT_MODEL_PREEDIT_STYLE_SELECTION = 6,
	TEXT_MODEL_PREEDIT_STYLE_INCORRECT = 7,
};
#endif /* TEXT_MODEL_PREEDIT_STYLE_ENUM */

/**
 * text_model - text model
 * @commit_string: commit
 * @preedit_string: pre-edit
 * @delete_surrounding_text: delete surrounding text
 * @cursor_position: set cursor to new position
 * @preedit_styling: pre-edit styling
 * @preedit_cursor: pre-edit cursor
 * @modifiers_map: modifiers map
 * @keysym: keysym
 * @enter: enter event
 * @leave: leave event
 * @input_panel_state: state of the input panel
 *
 * A model for text input. Adds support for text input and input methods
 * to applications. A text_model object is created from a
 * text_model_factory and corresponds typically to a text entry in an
 * application. Requests are used to activate/deactivate the model and set
 * information like surrounding and selected text or the content type. The
 * information about entered text is sent to the model via the pre-edit and
 * commit events. Using this interface removes the need for applications to
 * directly process hardware key events and compose text out of them.
 */
struct text_model_listener {
	/**
	 * commit_string - commit
	 * @serial: (none)
	 * @text: (none)
	 *
	 * Notify when text should be inserted into the editor widget.
	 * The text to commit could be either just a single character after
	 * a key press or the result of some composing (pre-edit). It also
	 * sets the new cursor position (as byte index) relative to the
	 * inserted text.
	 *
	 * Any previously set composing text should be removed.
	 */
	void (*commit_string)(void *data,
			      struct text_model *text_model,
			      uint32_t serial,
			      const char *text);
	/**
	 * preedit_string - pre-edit
	 * @serial: (none)
	 * @text: (none)
	 * @commit: (none)
	 *
	 * Notify when a new composing text (pre-edit) should be set
	 * around the current cursor position. Any previously set composing
	 * text should be removed.
	 *
	 * The commit text can be used to replace the preedit text on reset
	 * (for example on unfocus).
	 */
	void (*preedit_string)(void *data,
			       struct text_model *text_model,
			       uint32_t serial,
			       const char *text,
			       const char *commit);
	/**
	 * delete_surrounding_text - delete surrounding text
	 * @serial: (none)
	 * @index: (none)
	 * @length: (none)
	 *
	 * Notify when the text around the current cursor position should
	 * be deleted. Index is relative to the current cursor (as byte
	 * index). Length is the length of deleted text (as bytes).
	 */
	void (*delete_surrounding_text)(void *data,
					struct text_model *text_model,
					uint32_t serial,
					int32_t index,
					uint32_t length);
	/**
	 * cursor_position - set cursor to new position
	 * @serial: (none)
	 * @index: (none)
	 * @anchor: (none)
	 *
	 * Notify when the cursor or anchor position should be modified.
	 * It should take effect after the next commit_string event.
	 */
	void (*cursor_position)(void *data,
				struct text_model *text_model,
				uint32_t serial,
				int32_t index,
				int32_t anchor);
	/**
	 * preedit_styling - pre-edit styling
	 * @serial: (none)
	 * @index: (none)
	 * @length: (none)
	 * @style: (none)
	 *
	 * Sets styling information on composing text. The style is
	 * applied for length (in bytes) characters from index relative to
	 * the beginning of the composing text (as byte index). Multiple
	 * styles can be applied to a composing text.
	 *
	 * This event should be handled as part of a following
	 * preedit_string event.
	 */
	void (*preedit_styling)(void *data,
				struct text_model *text_model,
				uint32_t serial,
				uint32_t index,
				uint32_t length,
				uint32_t style);
	/**
	 * preedit_cursor - pre-edit cursor
	 * @serial: (none)
	 * @index: (none)
	 *
	 * Sets the cursor position inside the composing text (as byte
	 * index) relative to the start of the composing text.
	 *
	 * This event should be handled as part of a following
	 * preedit_string event.
	 */
	void (*preedit_cursor)(void *data,
			       struct text_model *text_model,
			       uint32_t serial,
			       int32_t index);
	/**
	 * modifiers_map - modifiers map
	 * @map: (none)
	 *
	 * Transfer an array of 0-terminated modifiers names. The
	 * position in the array is the index of the modifier as used in
	 * the modifiers bitmask in the keysym event.
	 */
	void (*modifiers_map)(void *data,
			      struct text_model *text_model,
			      struct wl_array *map);
	/**
	 * keysym - keysym
	 * @serial: (none)
	 * @time: (none)
	 * @sym: (none)
	 * @state: (none)
	 * @modifiers: (none)
	 *
	 * Notify when a key event was sent. Key events should not be
	 * used for normal text input operations, which should be done with
	 * commit_string, delete_surrounfing_text, etc. The key event
	 * follows the wl_keyboard key event convention. State is a XKB
	 * keysym, state a wl_keyboard key_state. Modifiers are a mask for
	 * effective modifiers (where the modfier indices are set by the
	 * modifiers_map event)
	 */
	void (*keysym)(void *data,
		       struct text_model *text_model,
		       uint32_t serial,
		       uint32_t time,
		       uint32_t sym,
		       uint32_t state,
		       uint32_t modifiers);
	/**
	 * enter - enter event
	 * @surface: (none)
	 *
	 * Notify the model when it is activated. Typically in response
	 * to an activate request.
	 */
	void (*enter)(void *data,
		      struct text_model *text_model,
		      struct wl_surface *surface);
	/**
	 * leave - leave event
	 *
	 * Notify the model when it is deactivated. Either in response to
	 * a deactivate request or when the assigned surface lost focus or
	 * was destroyed.
	 */
	void (*leave)(void *data,
		      struct text_model *text_model);
	/**
	 * input_panel_state - state of the input panel
	 * @state: (none)
	 *
	 * Notify when the visibility state of the input panel changed.
	 */
	void (*input_panel_state)(void *data,
				  struct text_model *text_model,
				  uint32_t state);
};

static inline int
text_model_add_listener(struct text_model *text_model,
			const struct text_model_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) text_model,
				     (void (**)(void)) listener, data);
}

#define TEXT_MODEL_SET_SURROUNDING_TEXT	0
#define TEXT_MODEL_ACTIVATE	1
#define TEXT_MODEL_DEACTIVATE	2
#define TEXT_MODEL_RESET	3
#define TEXT_MODEL_SET_CURSOR_RECTANGLE	4
#define TEXT_MODEL_SET_CONTENT_TYPE	5
#define TEXT_MODEL_INVOKE_ACTION	6
#define TEXT_MODEL_COMMIT	7
#define TEXT_MODEL_SHOW_INPUT_PANEL	8
#define TEXT_MODEL_HIDE_INPUT_PANEL	9

static inline void
text_model_set_user_data(struct text_model *text_model, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) text_model, user_data);
}

static inline void *
text_model_get_user_data(struct text_model *text_model)
{
	return wl_proxy_get_user_data((struct wl_proxy *) text_model);
}

static inline void
text_model_destroy(struct text_model *text_model)
{
	wl_proxy_destroy((struct wl_proxy *) text_model);
}

static inline void
text_model_set_surrounding_text(struct text_model *text_model, const char *text, uint32_t cursor, uint32_t anchor)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_SET_SURROUNDING_TEXT, text, cursor, anchor);
}

static inline void
text_model_activate(struct text_model *text_model, uint32_t serial, struct wl_seat *seat, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_ACTIVATE, serial, seat, surface);
}

static inline void
text_model_deactivate(struct text_model *text_model, struct wl_seat *seat)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_DEACTIVATE, seat);
}

static inline void
text_model_reset(struct text_model *text_model, uint32_t serial)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_RESET, serial);
}

static inline void
text_model_set_cursor_rectangle(struct text_model *text_model, int32_t x, int32_t y, int32_t width, int32_t height)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_SET_CURSOR_RECTANGLE, x, y, width, height);
}

static inline void
text_model_set_content_type(struct text_model *text_model, uint32_t hint, uint32_t purpose)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_SET_CONTENT_TYPE, hint, purpose);
}

static inline void
text_model_invoke_action(struct text_model *text_model, uint32_t button, uint32_t index)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_INVOKE_ACTION, button, index);
}

static inline void
text_model_commit(struct text_model *text_model)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_COMMIT);
}

static inline void
text_model_show_input_panel(struct text_model *text_model)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_SHOW_INPUT_PANEL);
}

static inline void
text_model_hide_input_panel(struct text_model *text_model)
{
	wl_proxy_marshal((struct wl_proxy *) text_model,
			 TEXT_MODEL_HIDE_INPUT_PANEL);
}

#define TEXT_MODEL_FACTORY_CREATE_TEXT_MODEL	0

static inline void
text_model_factory_set_user_data(struct text_model_factory *text_model_factory, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) text_model_factory, user_data);
}

static inline void *
text_model_factory_get_user_data(struct text_model_factory *text_model_factory)
{
	return wl_proxy_get_user_data((struct wl_proxy *) text_model_factory);
}

static inline void
text_model_factory_destroy(struct text_model_factory *text_model_factory)
{
	wl_proxy_destroy((struct wl_proxy *) text_model_factory);
}

static inline struct text_model *
text_model_factory_create_text_model(struct text_model_factory *text_model_factory)
{
	struct wl_proxy *id;

	id = wl_proxy_create((struct wl_proxy *) text_model_factory,
			     &text_model_interface);
	if (!id)
		return NULL;

	wl_proxy_marshal((struct wl_proxy *) text_model_factory,
			 TEXT_MODEL_FACTORY_CREATE_TEXT_MODEL, id);

	return (struct text_model *) id;
}

#ifdef  __cplusplus
}
#endif

#endif
