/* 
 * Copyright © 2012 Intel Corporation
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef INPUT_METHOD_SERVER_PROTOCOL_H
#define INPUT_METHOD_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-util.h"

struct wl_client;
struct wl_resource;

struct input_method_context;
struct input_method;
struct input_panel;
struct input_panel_surface;

extern const struct wl_interface input_method_context_interface;
extern const struct wl_interface input_method_interface;
extern const struct wl_interface input_panel_interface;
extern const struct wl_interface input_panel_surface_interface;

/**
 * input_method_context - input method context
 * @destroy: (none)
 * @commit_string: commit string
 * @preedit_string: pre-edit string
 * @preedit_styling: (none)
 * @preedit_cursor: pre-edit cursor
 * @delete_surrounding_text: (none)
 * @cursor_position: (none)
 * @modifiers_map: (none)
 * @keysym: keysym
 * @grab_keyboard: grab hardware keyboard
 * @key: (none)
 * @modifiers: (none)
 *
 * Corresponds to a text model on input method side. An input method
 * context is created on text mode activation on the input method side. It
 * allows to receive information about the text model from the application
 * via events. Input method contexts do not keep state after deactivation
 * and should be destroyed after deactivation is handled.
 */
struct input_method_context_interface {
	/**
	 * destroy - (none)
	 */
	void (*destroy)(struct wl_client *client,
			struct wl_resource *resource);
	/**
	 * commit_string - commit string
	 * @serial: (none)
	 * @text: (none)
	 *
	 * Send the commit string text to the applications text model.
	 */
	void (*commit_string)(struct wl_client *client,
			      struct wl_resource *resource,
			      uint32_t serial,
			      const char *text);
	/**
	 * preedit_string - pre-edit string
	 * @serial: (none)
	 * @text: (none)
	 * @commit: (none)
	 *
	 * Send the pre-edit string text to the applications text model.
	 */
	void (*preedit_string)(struct wl_client *client,
			       struct wl_resource *resource,
			       uint32_t serial,
			       const char *text,
			       const char *commit);
	/**
	 * preedit_styling - (none)
	 * @serial: (none)
	 * @index: (none)
	 * @length: (none)
	 * @style: (none)
	 */
	void (*preedit_styling)(struct wl_client *client,
				struct wl_resource *resource,
				uint32_t serial,
				uint32_t index,
				uint32_t length,
				uint32_t style);
	/**
	 * preedit_cursor - pre-edit cursor
	 * @serial: (none)
	 * @index: (none)
	 *
	 * Sets the cursor position inside the composing text (as byte
	 * index) relative to the start of the composing text.
	 */
	void (*preedit_cursor)(struct wl_client *client,
			       struct wl_resource *resource,
			       uint32_t serial,
			       int32_t index);
	/**
	 * delete_surrounding_text - (none)
	 * @serial: (none)
	 * @index: (none)
	 * @length: (none)
	 */
	void (*delete_surrounding_text)(struct wl_client *client,
					struct wl_resource *resource,
					uint32_t serial,
					int32_t index,
					uint32_t length);
	/**
	 * cursor_position - (none)
	 * @serial: (none)
	 * @index: (none)
	 * @anchor: (none)
	 */
	void (*cursor_position)(struct wl_client *client,
				struct wl_resource *resource,
				uint32_t serial,
				int32_t index,
				int32_t anchor);
	/**
	 * modifiers_map - (none)
	 * @map: (none)
	 */
	void (*modifiers_map)(struct wl_client *client,
			      struct wl_resource *resource,
			      struct wl_array *map);
	/**
	 * keysym - keysym
	 * @serial: (none)
	 * @time: (none)
	 * @sym: (none)
	 * @state: (none)
	 * @modifiers: (none)
	 *
	 * Notify when a key event was sent. Key events should not be
	 * used for normal text input operations, which should be done with
	 * commit_string, delete_surrounfing_text, etc. The key event
	 * follows the wl_keyboard key event convention. State is a XKB
	 * keysym, state a wl_keyboard key_state.
	 */
	void (*keysym)(struct wl_client *client,
		       struct wl_resource *resource,
		       uint32_t serial,
		       uint32_t time,
		       uint32_t sym,
		       uint32_t state,
		       uint32_t modifiers);
	/**
	 * grab_keyboard - grab hardware keyboard
	 * @keyboard: (none)
	 *
	 * Allows an input method to receive hardware keyboard input and
	 * process key events to generate text events (with pre-edit) over
	 * the. This allows input methods which compose multiple key events
	 * for inputting text like it is done for CJK languages.
	 */
	void (*grab_keyboard)(struct wl_client *client,
			      struct wl_resource *resource,
			      uint32_t keyboard);
	/**
	 * key - (none)
	 * @serial: (none)
	 * @time: (none)
	 * @key: (none)
	 * @state: (none)
	 */
	void (*key)(struct wl_client *client,
		    struct wl_resource *resource,
		    uint32_t serial,
		    uint32_t time,
		    uint32_t key,
		    uint32_t state);
	/**
	 * modifiers - (none)
	 * @serial: (none)
	 * @mods_depressed: (none)
	 * @mods_latched: (none)
	 * @mods_locked: (none)
	 * @group: (none)
	 */
	void (*modifiers)(struct wl_client *client,
			  struct wl_resource *resource,
			  uint32_t serial,
			  uint32_t mods_depressed,
			  uint32_t mods_latched,
			  uint32_t mods_locked,
			  uint32_t group);
};

#define INPUT_METHOD_CONTEXT_SURROUNDING_TEXT	0
#define INPUT_METHOD_CONTEXT_RESET	1
#define INPUT_METHOD_CONTEXT_CONTENT_TYPE	2
#define INPUT_METHOD_CONTEXT_INVOKE_ACTION	3
#define INPUT_METHOD_CONTEXT_COMMIT	4

static inline void
input_method_context_send_surrounding_text(struct wl_resource *resource_, const char *text, uint32_t cursor, uint32_t anchor)
{
	wl_resource_post_event(resource_, INPUT_METHOD_CONTEXT_SURROUNDING_TEXT, text, cursor, anchor);
}

static inline void
input_method_context_send_reset(struct wl_resource *resource_, uint32_t serial)
{
	wl_resource_post_event(resource_, INPUT_METHOD_CONTEXT_RESET, serial);
}

static inline void
input_method_context_send_content_type(struct wl_resource *resource_, uint32_t hint, uint32_t purpose)
{
	wl_resource_post_event(resource_, INPUT_METHOD_CONTEXT_CONTENT_TYPE, hint, purpose);
}

static inline void
input_method_context_send_invoke_action(struct wl_resource *resource_, uint32_t button, uint32_t index)
{
	wl_resource_post_event(resource_, INPUT_METHOD_CONTEXT_INVOKE_ACTION, button, index);
}

static inline void
input_method_context_send_commit(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, INPUT_METHOD_CONTEXT_COMMIT);
}

#define INPUT_METHOD_ACTIVATE	0
#define INPUT_METHOD_DEACTIVATE	1

static inline void
input_method_send_activate(struct wl_resource *resource_, struct wl_resource *id, uint32_t serial)
{
	wl_resource_post_event(resource_, INPUT_METHOD_ACTIVATE, id, serial);
}

static inline void
input_method_send_deactivate(struct wl_resource *resource_, struct wl_resource *context)
{
	wl_resource_post_event(resource_, INPUT_METHOD_DEACTIVATE, context);
}

/**
 * input_panel - interface for implementing keyboards
 * @get_input_panel_surface: (none)
 *
 * Only one client can bind this interface at a time.
 */
struct input_panel_interface {
	/**
	 * get_input_panel_surface - (none)
	 * @id: (none)
	 * @surface: (none)
	 */
	void (*get_input_panel_surface)(struct wl_client *client,
					struct wl_resource *resource,
					uint32_t id,
					struct wl_resource *surface);
};

#ifndef INPUT_PANEL_SURFACE_POSITION_ENUM
#define INPUT_PANEL_SURFACE_POSITION_ENUM
enum input_panel_surface_position {
	INPUT_PANEL_SURFACE_POSITION_CENTER_BOTTOM = 0,
};
#endif /* INPUT_PANEL_SURFACE_POSITION_ENUM */

struct input_panel_surface_interface {
	/**
	 * set_toplevel - set the surface type as a keyboard
	 * @position: (none)
	 *
	 * A keybaord surface is only shown, when a text model is active
	 */
	void (*set_toplevel)(struct wl_client *client,
			     struct wl_resource *resource,
			     uint32_t position);
};

#ifdef  __cplusplus
}
#endif

#endif
