/* 
 * Copyright © 2012 Intel Corporation
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef TEXT_SERVER_PROTOCOL_H
#define TEXT_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-util.h"

struct wl_client;
struct wl_resource;

struct text_model;
struct text_model_factory;

extern const struct wl_interface text_model_interface;
extern const struct wl_interface text_model_factory_interface;

#ifndef TEXT_MODEL_CONTENT_HINT_ENUM
#define TEXT_MODEL_CONTENT_HINT_ENUM
/**
 * text_model_content_hint - content hint
 * @TEXT_MODEL_CONTENT_HINT_NONE: no special behaviour
 * @TEXT_MODEL_CONTENT_HINT_DEFAULT: auto completion, correction and
 *	capitalization
 * @TEXT_MODEL_CONTENT_HINT_PASSWORD: hidden and sensitive text
 * @TEXT_MODEL_CONTENT_HINT_AUTO_COMPLETION: suggest word completions
 * @TEXT_MODEL_CONTENT_HINT_AUTO_CORRECTION: suggest word corrections
 * @TEXT_MODEL_CONTENT_HINT_AUTO_CAPITALIZATION: switch to uppercase
 *	letters at the start of a sentence
 * @TEXT_MODEL_CONTENT_HINT_LOWERCASE: prefer lowercase letters
 * @TEXT_MODEL_CONTENT_HINT_UPPERCASE: prefer uppercase letters
 * @TEXT_MODEL_CONTENT_HINT_TITLECASE: prefer casing for titles and
 *	headings (can be language dependend)
 * @TEXT_MODEL_CONTENT_HINT_HIDDEN_TEXT: characters should be hidden
 * @TEXT_MODEL_CONTENT_HINT_SENSITIVE_DATA: typed text should not be
 *	stored
 * @TEXT_MODEL_CONTENT_HINT_LATIN: just latin characters should be
 *	entered
 * @TEXT_MODEL_CONTENT_HINT_MULTILINE: the text input is multiline
 *
 * Content hint is a bitmask to allow to modify the behavior of the text
 * input
 */
enum text_model_content_hint {
	TEXT_MODEL_CONTENT_HINT_NONE = 0x0,
	TEXT_MODEL_CONTENT_HINT_DEFAULT = 0x7,
	TEXT_MODEL_CONTENT_HINT_PASSWORD = 0xc,
	TEXT_MODEL_CONTENT_HINT_AUTO_COMPLETION = 0x1,
	TEXT_MODEL_CONTENT_HINT_AUTO_CORRECTION = 0x2,
	TEXT_MODEL_CONTENT_HINT_AUTO_CAPITALIZATION = 0x4,
	TEXT_MODEL_CONTENT_HINT_LOWERCASE = 0x8,
	TEXT_MODEL_CONTENT_HINT_UPPERCASE = 0x10,
	TEXT_MODEL_CONTENT_HINT_TITLECASE = 0x20,
	TEXT_MODEL_CONTENT_HINT_HIDDEN_TEXT = 0x40,
	TEXT_MODEL_CONTENT_HINT_SENSITIVE_DATA = 0x80,
	TEXT_MODEL_CONTENT_HINT_LATIN = 0x100,
	TEXT_MODEL_CONTENT_HINT_MULTILINE = 0x200,
};
#endif /* TEXT_MODEL_CONTENT_HINT_ENUM */

#ifndef TEXT_MODEL_CONTENT_PURPOSE_ENUM
#define TEXT_MODEL_CONTENT_PURPOSE_ENUM
/**
 * text_model_content_purpose - content purpose
 * @TEXT_MODEL_CONTENT_PURPOSE_NORMAL: default input, allowing all
 *	characters
 * @TEXT_MODEL_CONTENT_PURPOSE_ALPHA: allow only alphabetic characters
 * @TEXT_MODEL_CONTENT_PURPOSE_DIGITS: allow only digits
 * @TEXT_MODEL_CONTENT_PURPOSE_NUMBER: input a number (including decimal
 *	separator and sign)
 * @TEXT_MODEL_CONTENT_PURPOSE_PHONE: input a phone number
 * @TEXT_MODEL_CONTENT_PURPOSE_URL: input an URL
 * @TEXT_MODEL_CONTENT_PURPOSE_EMAIL: input an email address
 * @TEXT_MODEL_CONTENT_PURPOSE_NAME: input a name of a person
 * @TEXT_MODEL_CONTENT_PURPOSE_PASSWORD: input a password (combine with
 *	password or sensitive_data hint)
 * @TEXT_MODEL_CONTENT_PURPOSE_DATE: input a date
 * @TEXT_MODEL_CONTENT_PURPOSE_TIME: input a time
 * @TEXT_MODEL_CONTENT_PURPOSE_DATETIME: input a date and time
 * @TEXT_MODEL_CONTENT_PURPOSE_TERMINAL: input for a terminal
 *
 * The content purpose allows to specify the primary purpose of a text
 * input.
 *
 * This allows an input method to show special purpose input panels with
 * extra characters or to disallow some characters.
 */
enum text_model_content_purpose {
	TEXT_MODEL_CONTENT_PURPOSE_NORMAL = 0,
	TEXT_MODEL_CONTENT_PURPOSE_ALPHA = 1,
	TEXT_MODEL_CONTENT_PURPOSE_DIGITS = 2,
	TEXT_MODEL_CONTENT_PURPOSE_NUMBER = 3,
	TEXT_MODEL_CONTENT_PURPOSE_PHONE = 4,
	TEXT_MODEL_CONTENT_PURPOSE_URL = 5,
	TEXT_MODEL_CONTENT_PURPOSE_EMAIL = 6,
	TEXT_MODEL_CONTENT_PURPOSE_NAME = 7,
	TEXT_MODEL_CONTENT_PURPOSE_PASSWORD = 8,
	TEXT_MODEL_CONTENT_PURPOSE_DATE = 9,
	TEXT_MODEL_CONTENT_PURPOSE_TIME = 10,
	TEXT_MODEL_CONTENT_PURPOSE_DATETIME = 11,
	TEXT_MODEL_CONTENT_PURPOSE_TERMINAL = 12,
};
#endif /* TEXT_MODEL_CONTENT_PURPOSE_ENUM */

#ifndef TEXT_MODEL_PREEDIT_STYLE_ENUM
#define TEXT_MODEL_PREEDIT_STYLE_ENUM
enum text_model_preedit_style {
	TEXT_MODEL_PREEDIT_STYLE_DEFAULT = 1,
	TEXT_MODEL_PREEDIT_STYLE_ACTIVE = 2,
	TEXT_MODEL_PREEDIT_STYLE_INACTIVE = 3,
	TEXT_MODEL_PREEDIT_STYLE_HIGHLIGHT = 4,
	TEXT_MODEL_PREEDIT_STYLE_UNDERLINE = 5,
	TEXT_MODEL_PREEDIT_STYLE_SELECTION = 6,
	TEXT_MODEL_PREEDIT_STYLE_INCORRECT = 7,
};
#endif /* TEXT_MODEL_PREEDIT_STYLE_ENUM */

/**
 * text_model - text model
 * @set_surrounding_text: sets the surrounding text
 * @activate: request activation
 * @deactivate: request deactivation
 * @reset: reset
 * @set_cursor_rectangle: (none)
 * @set_content_type: set content purpose and hint
 * @invoke_action: (none)
 * @commit: (none)
 * @show_input_panel: show input panels
 * @hide_input_panel: hide input panels
 *
 * A model for text input. Adds support for text input and input methods
 * to applications. A text_model object is created from a
 * text_model_factory and corresponds typically to a text entry in an
 * application. Requests are used to activate/deactivate the model and set
 * information like surrounding and selected text or the content type. The
 * information about entered text is sent to the model via the pre-edit and
 * commit events. Using this interface removes the need for applications to
 * directly process hardware key events and compose text out of them.
 */
struct text_model_interface {
	/**
	 * set_surrounding_text - sets the surrounding text
	 * @text: (none)
	 * @cursor: (none)
	 * @anchor: (none)
	 *
	 * Sets the plain surrounding text around the input position.
	 * Cursor is the byte index within the surrounding text. Anchor is
	 * the byte index of the selection anchor within the surrounding
	 * text. If there is no selected text anchor is the same as cursor.
	 */
	void (*set_surrounding_text)(struct wl_client *client,
				     struct wl_resource *resource,
				     const char *text,
				     uint32_t cursor,
				     uint32_t anchor);
	/**
	 * activate - request activation
	 * @serial: (none)
	 * @seat: (none)
	 * @surface: (none)
	 *
	 * Requests the model to be activated (typically when the text
	 * entry gets focus). The seat argument is a wl_seat which
	 * maintains the focus for this activation. The surface argument is
	 * a wl_surface assigned to the model and tracked for focus lost.
	 * The activated event is emitted on successful activation.
	 */
	void (*activate)(struct wl_client *client,
			 struct wl_resource *resource,
			 uint32_t serial,
			 struct wl_resource *seat,
			 struct wl_resource *surface);
	/**
	 * deactivate - request deactivation
	 * @seat: (none)
	 *
	 * Requests the model to be deactivated (typically when the text
	 * entry lost focus). The seat argument is a wl_seat which was used
	 * for activation.
	 */
	void (*deactivate)(struct wl_client *client,
			   struct wl_resource *resource,
			   struct wl_resource *seat);
	/**
	 * reset - reset
	 * @serial: (none)
	 *
	 * Should be called by an editor widget when the input state
	 * should be reseted, for example after the text was changed
	 * outside of the normal input method flow.
	 */
	void (*reset)(struct wl_client *client,
		      struct wl_resource *resource,
		      uint32_t serial);
	/**
	 * set_cursor_rectangle - (none)
	 * @x: (none)
	 * @y: (none)
	 * @width: (none)
	 * @height: (none)
	 */
	void (*set_cursor_rectangle)(struct wl_client *client,
				     struct wl_resource *resource,
				     int32_t x,
				     int32_t y,
				     int32_t width,
				     int32_t height);
	/**
	 * set_content_type - set content purpose and hint
	 * @hint: (none)
	 * @purpose: (none)
	 *
	 * Sets the content purpose and content hint. While the purpose
	 * is the basic purpose of an input field, the hint flags allow to
	 * modify some of the behavior.
	 *
	 * When no content type is explicitly set, a normal content purpose
	 * with default hints (auto completion, auto correction, auto
	 * capitalization) should be assumed.
	 */
	void (*set_content_type)(struct wl_client *client,
				 struct wl_resource *resource,
				 uint32_t hint,
				 uint32_t purpose);
	/**
	 * invoke_action - (none)
	 * @button: (none)
	 * @index: (none)
	 */
	void (*invoke_action)(struct wl_client *client,
			      struct wl_resource *resource,
			      uint32_t button,
			      uint32_t index);
	/**
	 * commit - (none)
	 */
	void (*commit)(struct wl_client *client,
		       struct wl_resource *resource);
	/**
	 * show_input_panel - show input panels
	 *
	 * Requests input panels (virtual keyboard) to show.
	 */
	void (*show_input_panel)(struct wl_client *client,
				 struct wl_resource *resource);
	/**
	 * hide_input_panel - hide input panels
	 *
	 * Requests input panels (virtual keyboard) to hide.
	 */
	void (*hide_input_panel)(struct wl_client *client,
				 struct wl_resource *resource);
};

#define TEXT_MODEL_COMMIT_STRING	0
#define TEXT_MODEL_PREEDIT_STRING	1
#define TEXT_MODEL_DELETE_SURROUNDING_TEXT	2
#define TEXT_MODEL_CURSOR_POSITION	3
#define TEXT_MODEL_PREEDIT_STYLING	4
#define TEXT_MODEL_PREEDIT_CURSOR	5
#define TEXT_MODEL_MODIFIERS_MAP	6
#define TEXT_MODEL_KEYSYM	7
#define TEXT_MODEL_ENTER	8
#define TEXT_MODEL_LEAVE	9
#define TEXT_MODEL_INPUT_PANEL_STATE	10

static inline void
text_model_send_commit_string(struct wl_resource *resource_, uint32_t serial, const char *text)
{
	wl_resource_post_event(resource_, TEXT_MODEL_COMMIT_STRING, serial, text);
}

static inline void
text_model_send_preedit_string(struct wl_resource *resource_, uint32_t serial, const char *text, const char *commit)
{
	wl_resource_post_event(resource_, TEXT_MODEL_PREEDIT_STRING, serial, text, commit);
}

static inline void
text_model_send_delete_surrounding_text(struct wl_resource *resource_, uint32_t serial, int32_t index, uint32_t length)
{
	wl_resource_post_event(resource_, TEXT_MODEL_DELETE_SURROUNDING_TEXT, serial, index, length);
}

static inline void
text_model_send_cursor_position(struct wl_resource *resource_, uint32_t serial, int32_t index, int32_t anchor)
{
	wl_resource_post_event(resource_, TEXT_MODEL_CURSOR_POSITION, serial, index, anchor);
}

static inline void
text_model_send_preedit_styling(struct wl_resource *resource_, uint32_t serial, uint32_t index, uint32_t length, uint32_t style)
{
	wl_resource_post_event(resource_, TEXT_MODEL_PREEDIT_STYLING, serial, index, length, style);
}

static inline void
text_model_send_preedit_cursor(struct wl_resource *resource_, uint32_t serial, int32_t index)
{
	wl_resource_post_event(resource_, TEXT_MODEL_PREEDIT_CURSOR, serial, index);
}

static inline void
text_model_send_modifiers_map(struct wl_resource *resource_, struct wl_array *map)
{
	wl_resource_post_event(resource_, TEXT_MODEL_MODIFIERS_MAP, map);
}

static inline void
text_model_send_keysym(struct wl_resource *resource_, uint32_t serial, uint32_t time, uint32_t sym, uint32_t state, uint32_t modifiers)
{
	wl_resource_post_event(resource_, TEXT_MODEL_KEYSYM, serial, time, sym, state, modifiers);
}

static inline void
text_model_send_enter(struct wl_resource *resource_, struct wl_resource *surface)
{
	wl_resource_post_event(resource_, TEXT_MODEL_ENTER, surface);
}

static inline void
text_model_send_leave(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, TEXT_MODEL_LEAVE);
}

static inline void
text_model_send_input_panel_state(struct wl_resource *resource_, uint32_t state)
{
	wl_resource_post_event(resource_, TEXT_MODEL_INPUT_PANEL_STATE, state);
}

/**
 * text_model_factory - text model factory
 * @create_text_model: create text model
 *
 * A factory for text models. This object is a singleton global.
 */
struct text_model_factory_interface {
	/**
	 * create_text_model - create text model
	 * @id: (none)
	 *
	 * Creates a new text model object.
	 */
	void (*create_text_model)(struct wl_client *client,
				  struct wl_resource *resource,
				  uint32_t id);
};

#ifdef  __cplusplus
}
#endif

#endif
